
package gov.anl.aps.dm.portal.controllers;

import gov.anl.aps.dm.portal.exceptions.DmPortalException;
import gov.anl.aps.dm.portal.exceptions.ObjectAlreadyExists;
import gov.anl.aps.dm.portal.exceptions.InvalidDate;
import gov.anl.aps.dm.portal.exceptions.MissingProperty;
import gov.anl.aps.dm.portal.model.entities.Experiment;
import gov.anl.aps.dm.portal.model.beans.ExperimentFacade;
import gov.anl.aps.dm.portal.model.beans.ExperimentTypeFacade;
import gov.anl.aps.dm.portal.model.beans.RoleTypeFacade;
import gov.anl.aps.dm.portal.model.beans.UserExperimentRoleFacade;
import gov.anl.aps.dm.portal.model.beans.UserInfoFacade;
import gov.anl.aps.dm.portal.model.entities.CloneableEntity;
import gov.anl.aps.dm.portal.model.entities.RoleType;
import gov.anl.aps.dm.portal.model.entities.UserExperimentRole;
import gov.anl.aps.dm.portal.model.entities.UserInfo;
import gov.anl.aps.dm.portal.utilities.SessionUtility;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.ejb.EJB;
import javax.inject.Named;
import javax.enterprise.context.SessionScoped;
import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;
import javax.faces.convert.Converter;
import javax.faces.convert.FacesConverter;
import org.apache.log4j.Logger;

@Named("experimentController")
@SessionScoped
public class ExperimentController extends CrudEntityController<Experiment, ExperimentFacade>
{

    private static final Logger logger = Logger.getLogger(ExperimentController.class.getName());

    @EJB
    private ExperimentFacade experimentFacade;

    @EJB
    private ExperimentTypeFacade experimentTypeFacade;

    @EJB
    private UserInfoFacade userInfoFacade;

    @EJB
    private RoleTypeFacade roleTypeFacade;
    
    @EJB
    private UserExperimentRoleFacade userExperimentRoleFacade;
    
    public class ExperimentUser extends CloneableEntity {

    String username;
    String name;
    int userId;
    int experimentId;
    private final boolean [] roles;

    public ExperimentUser(int experimentId, int userId) {
        roles = new boolean[maxExperimentRoleTypeId];
        this.experimentId = experimentId;
        this.userId = userId;
    }
    
    public String getUsername() {
        return username;
    }

    public void setUsername(String username) {
        this.username = username;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public boolean isIsManager() {
        return roles[2];
    }

    public void setIsManager(boolean isManager) {
        logger.debug("setting Manager role for experiment "+ experimentId + " and user " + username + " to " + isManager);
        roles[2] = isManager;
    }

    public boolean isIsPI() {
        return roles[3];
    }

    public void setIsPI(boolean isPI) {
        roles[3] = isPI;
    }

    public boolean isIsUser() {
        return roles[4];
    }

    public void setIsUser(boolean isUser) {
        roles[4] = isUser;
    }

    public void setIsInRole(RoleType role, boolean isInRole) {
        roles[role.getId()] = isInRole;
    }
 
    public boolean[] getRoles() {
        return roles;
    }
    
    public void removeAllExperimentRoles() {
        UserExperimentRole userExperimentRole;
        for (int i = 1; i <= 4; i++) {
        if (roles[i]) {
                userExperimentRole = userExperimentRoleFacade.findByUserExperimentRole(userId, experimentId, i);
                userExperimentRoleFacade.remove(userExperimentRole);
            }            
        }       
    }
    
    public void removeExperimentRole(int roleId) {
        UserExperimentRole userExperimentRole = userExperimentRoleFacade.findByUserExperimentRole(userId, experimentId, roleId);
        userExperimentRoleFacade.remove(userExperimentRole);
    }

    public boolean canAddUsers() {
        return roles[2] || roles[3];            
    }
    
    @Override
    public int hashCode() {
        return userId + experimentId * 100;
    }

    @Override
    public boolean equals(Object object) {
        // TODO: Warning - this method won't work in the case the id fields are not set
        if (!(object instanceof ExperimentUser)) {
            return false;
        }
        return ((userId == ((ExperimentUser)object).userId) && (experimentId == ((ExperimentUser)object).experimentId));
    }

}
    
    class ExperimentUsersTable extends DataTableController<ExperimentUser> {

        @Override
        public String getClassName() {
            return "ExperimentUser";
        }

        @Override
        public List<ExperimentUser> findAll() {
            List<UserInfo> list = userInfoFacade.findUsersInExperiment(getCurrent().getId());
            return convertExperimentUsers(list);
//            return new ArrayList<>(tempUsersList);
        }
        
        List<ExperimentUser> convertExperimentUsers(List<UserInfo> list) {
            logger.debug("converting ExperimentUser ");
            List<RoleType> roleTypes = roleTypeFacade.findByRoleSystemType(false);
            ExperimentUser experimentUser;
            for (UserInfo user : list) {
                if ((experimentUser = experimentUsers.get(user)) == null) {
                    experimentUser = new ExperimentUser(getCurrent().getId(), user.getId());
                    experimentUser.setUsername(user.getUsername());
                    experimentUser.setName(user.getLastName() + ", " + user.getFirstName());
                    experimentUsers.put(user, experimentUser);
                }
                for (RoleType roleType : roleTypes) {
                    boolean inRole = user.inRole(roleType.getId(), getCurrent().getId());
                    experimentUser.setIsInRole(roleType, inRole);
                }
            }
            return new ArrayList<>(experimentUsers.values());
        }
        
        @Override
        public String getTableName() {
            return "experimentUsersTable";
        }
    }

    class NoExperimentUsersTypeTable extends DataTableController<UserInfo> {

        @Override
        public String getClassName() {
            return "UserInfo";
        }

        @Override
        public List<UserInfo> findAll() {
            resetList();
            return userInfoFacade.findNoUsersInExperiment(getCurrent().getId());
        }
        
        @Override
        public String getTableName() {
            return "noExperimentUsersTable";
        }
    }
    
    private ExperimentController.ExperimentUsersTable experimentUsersListTable;
    private ExperimentController.ExperimentUsersTable experimentUsersEditTable;
    private ExperimentController.NoExperimentUsersTypeTable noExperimentUsersTypeTable;
    final private Map<UserInfo, ExperimentUser> experimentUsers = new HashMap<>();
    int maxExperimentRoleTypeId = 5;

    public ExperimentController() {
//        List<RoleType> roleTypesList = roleTypeFacade.findAll();
//        for (RoleType roleType : roleTypesList) {
//            if ((roleType.getId() > maxExperimentRoleTypeId) && !roleType.isIsSystemRole()) {
//                maxExperimentRoleTypeId = roleType.getId();
//            }
//        }
        this.experimentUsersListTable = new ExperimentController.ExperimentUsersTable();
        this.experimentUsersEditTable = new ExperimentController.ExperimentUsersTable();
        this.noExperimentUsersTypeTable = new ExperimentController.NoExperimentUsersTypeTable();
    }

    @Override
    protected ExperimentFacade getFacade() {
        return experimentFacade;
    }

    @Override
    protected Experiment createEntityInstance() {
        return new Experiment();
    }

    @Override
    public String getEntityTypeName() {
        return "experiment";
    }

    @Override
    public String getCurrentEntityInstanceName() {
        if (getCurrent() != null) {
            return getCurrent().getName();
        }
        return "";
    }

    @Override
    public List<Experiment> getAvailableItems() {
        return super.getAvailableItems();
    }

    @Override
    public String prepareEdit(Experiment experiment) {
        clear();
        return super.prepareEdit(experiment);
    }

    @Override
    public String prepareView(Experiment experiment) {
        clear();
        return super.prepareView(experiment);
    }

    @Override
    public void prepareEntityInsert(Experiment experiment) throws ObjectAlreadyExists, InvalidDate, MissingProperty {
        if ((experiment.getName() == null) || (experiment.getName().length() == 0)){
            throw new MissingProperty("Experiment name is missing.");
        }
        Experiment existingExperiment = experimentFacade.findByName(experiment.getName());
        if (existingExperiment != null) {
            throw new ObjectAlreadyExists("Experiment " + experiment.getName() + " already exists.");
        }

        verifyExperiment(experiment);
        logger.debug("Inserting new experiment " + experiment.getName());
    }

    @Override
    public void prepareEntityUpdate(Experiment experiment) throws DmPortalException {
        if ((experiment.getName() == null) || (experiment.getName().length() == 0)) {
            throw new MissingProperty("Experiment name is missing.");
        }
        verifyExperiment(experiment);        
        logger.debug("Updating experiment " + experiment.getName());                
    }

     @Override
     protected String getObjectAlreadyExistMessage(Experiment experiment)  {
        if (experiment == null) {
            return null;
        }
        return "Experiment " + experiment.getName() + " already exists."; 
    }
     
    private void verifyExperiment(Experiment experiment) throws InvalidDate, MissingProperty {       
        if (experiment.getExperimentType() == null) {
            throw new MissingProperty("Experiment Type is missing.");
        }
        
        if ((experiment.getStartDate() != null) && (experiment.getEndDate() != null) && (experiment.getEndDate().before(experiment.getStartDate()))) {
            throw new InvalidDate("Experiment end date is before start date.");
        }
    }
   
// the methods for Experiment User

    public void clear() {
        experimentUsersListTable.resetList();
        experimentUsersEditTable.resetList();
        experimentUsers.clear();
        noExperimentUsersTypeTable.resetList();
    }
    
    public ExperimentController.ExperimentUsersTable getExperimentUsersListTable() {
        return experimentUsersListTable;
    }

    public void setExperimentUsersListTable(ExperimentController.ExperimentUsersTable experimentUsersTable) {
        this.experimentUsersListTable = experimentUsersTable;
    }
    
    public ExperimentController.ExperimentUsersTable getExperimentUsersEditTable() {
        return experimentUsersEditTable;
    }

    public void setExperimentUsersEditTable(ExperimentController.ExperimentUsersTable experimentUsersTable) {
        this.experimentUsersEditTable = experimentUsersTable;
    }
    
    public ExperimentController.NoExperimentUsersTypeTable getNoExperimentUsersTypeTable() {
        return noExperimentUsersTypeTable;
    }

    public void setNoExperimentUsersTypeTable(ExperimentController.NoExperimentUsersTypeTable noExperimentUsersTypeTable) {
        this.noExperimentUsersTypeTable = noExperimentUsersTypeTable;
    }
    
    public boolean canAddUsers(String username) {
        UserInfo user = userInfoFacade.findByUsername(username);
        // optimally here it should find all system roles. The roles optimally would have a policy
        // defining adding/removing users. For each of the role that allows adding users,
        // the code that is for administartor role (below) would be added.
        RoleType rt = roleTypeFacade.findById(1);   // administrator role id
        if (rt == null) {
            return false;
        }
        if (rt.isAdmin(username)) {
            return true;
        }
        return false;
        // experiment type permission
        // return experimentUsersMap.get(username).canAddUsers();
    }

     @Override
    public String update() {
        boolean updated;
        List<RoleType> roleTypesList = roleTypeFacade.findAll();
        for (UserInfo user : experimentUsers.keySet()) {
            updated = false;
            ExperimentUser experimentUser = experimentUsers.get(user);
            for (RoleType roleType : roleTypesList) {
                if (!roleType.isIsSystemRole()) {
                    int roleId = roleType.getId();
                    if (!user.inRole(roleId, experimentUser.experimentId)) {
                        if (experimentUser.roles[roleId]) {
                            UserExperimentRole userExperimentRole = new UserExperimentRole(experimentUser.userId, experimentUser.experimentId, roleId);
                            user.addUserExperimentRole(userExperimentRole);
                            updated = true;
                            logger.debug("adding userExperimentRole " + experimentUser.userId + " " + experimentUser.experimentId + " " + roleId);
                        }
                    }
                    else if (!experimentUser.roles[roleId]) {
                            user.removeExperimentRole(roleId, experimentUser.experimentId);
                            updated = true;
                            logger.debug("adding userExperimentRole " + experimentUser.userId + " " + experimentUser.experimentId + " " + roleId);
                        }
                    }
                    if (updated) {
                        userInfoFacade.edit(user);
                    }
            }
        }
    
        clear();
//        experimentUsersListTable.resetList();
//        experimentUsersEditTable.resetList();
        return super.update();
    }

   public String getRemovedUserName() {
        if (experimentUsersEditTable.getCurrentObject() != null) {
            return experimentUsersEditTable.getCurrentObject().getUsername();
        }
        return "";
    }
        
    public String removeUser() {

        List<RoleType> roleTypesList = roleTypeFacade.findAll();
        ExperimentUser experimentUser = experimentUsersEditTable.getCurrentObject();
        UserInfo user = findUserInExperiment(experimentUser);
        for (RoleType roleType : roleTypesList) {
            int roleId = roleType.getId();
            if (experimentUser.roles[roleId]) {
                user.removeExperimentRole(roleId, experimentUser.experimentId);
            }
        }
        userInfoFacade.edit(user);       
        clear();
        return "edit?faces-redirect=true";
    }
    
    public void removeUserRoles() {
        ExperimentUser experimentUser = experimentUsersEditTable.getCurrentObject();
        experimentUser.removeAllExperimentRoles();        
    }
    
    public void updateRemovedExperimentRoles() {
        List<RoleType> roleTypesList = roleTypeFacade.findAll();
        for (UserInfo user : experimentUsers.keySet()) {
            ExperimentUser experimentUser = experimentUsers.get(user);
            for (RoleType roleType : roleTypesList) {
                int roleId = roleType.getId();
                if (!roleType.isIsSystemRole() && (!experimentUser.roles[roleId]) && (user.inRole(roleId, experimentUser.experimentId))) {
                        experimentUser.removeExperimentRole(roleId);
                }
            }
        }
    }

    private UserInfo findUserInExperiment(ExperimentUser experimentUser) {
        for (UserInfo user : experimentUsers.keySet()) {
            if (experimentUsers.get(user).equals(experimentUser)) {
                return user;
            }
        }
        return null;
    }

    public String addExperimentUser() {
        List<UserInfo> newUsers = noExperimentUsersTypeTable.getSelectedObjectList();
        if (newUsers == null) {
            logger.debug("null selected list");
            return null;
        } else if (newUsers.isEmpty()) {
            logger.debug("empty selected list");
            return null;
        } else {
            try {
                UserExperimentRole userExperimentRole;
                for (UserInfo user : newUsers) {
                    userExperimentRole = new UserExperimentRole(user.getId(), current.getId(), 4); // set the User
                    user.getUserExperimentRoleList().add(userExperimentRole);
                    userInfoFacade.edit(user);
                }
                clear();
                return "edit?faces-redirect=true";
            } catch (RuntimeException ex) {
                SessionUtility.addErrorMessage("Error", "Could not update UserInfo" + ": " + ex.getMessage());
                return null;
            }
        }
    }

    @FacesConverter(forClass = Experiment.class)
    public static class ExperimentControllerConverter implements Converter {

        @Override
        public Object getAsObject(FacesContext facesContext, UIComponent component, String value) {
            if (value == null || value.length() == 0) {
                return null;
            }
            ExperimentController controller = (ExperimentController) facesContext.getApplication().getELResolver().
                    getValue(facesContext.getELContext(), null, "experimentController");
            return controller.getEntity(getKey(value));
        }

        java.lang.Integer getKey(String value) {
            java.lang.Integer key;
            key = Integer.valueOf(value);
            return key;
        }

        String getStringKey(java.lang.Integer value) {
            StringBuilder sb = new StringBuilder();
            sb.append(value);
            return sb.toString();
        }

        @Override
        public String getAsString(FacesContext facesContext, UIComponent component, Object object) {
            if (object == null) {
                return null;
            }
            if (object instanceof Experiment) {
                Experiment o = (Experiment) object;
                return getStringKey(o.getId());
            }
            else {
                throw new IllegalArgumentException("object " + object + " is of type " + object.getClass().getName() + "; expected type: " + Experiment.class.getName());
            }
        }

    }

}